// Copyright 2016-2017, Pulumi Corporation.  All rights reserved.

package graph

import (
	"github.com/pkg/errors"
)

// Topsort topologically sorts the graph, yielding an array of nodes that are in dependency order, using a simple
// DFS-based algorithm.  The graph must be acyclic, otherwise this function will return an error.
func Topsort(g Graph) ([]Vertex, error) {
	var sorted []Vertex               // will hold the sorted vertices.
	visiting := make(map[Vertex]bool) // temporary entries to detect cycles.
	visited := make(map[Vertex]bool)  // entries to avoid visiting the same node twice.

	// Now enumerate the roots, topologically sorting their dependencies.
	roots := g.Roots()
	for _, r := range roots {
		if err := topvisit(r.To(), &sorted, visiting, visited); err != nil {
			return sorted, err
		}
	}
	return sorted, nil
}

func topvisit(n Vertex, sorted *[]Vertex, visiting map[Vertex]bool, visited map[Vertex]bool) error {
	if visiting[n] {
		// This is not a DAG!  Stop sorting right away, and issue an error.
		// IDEA: return diagnostic information about why this isn't a DAG (e.g., full cycle path).
		return errors.New("Graph is not a DAG")
	}
	if !visited[n] {
		visiting[n] = true
		for _, m := range n.Outs() {
			if err := topvisit(m.To(), sorted, visiting, visited); err != nil {
				return err
			}
		}
		visited[n] = true
		visiting[n] = false
		*sorted = append(*sorted, n)
	}
	return nil
}
