package integration

import (
	"bytes"
	"encoding/json"
	"fmt"
	"path"
	"time"

	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/aws/session"
	"github.com/aws/aws-sdk-go/service/s3"

	"github.com/pulumi/pulumi/pkg/resource"
)

// S3Reporter is a TestStatsReporter that publises test data to S3
type S3Reporter struct {
	s3svc     *s3.S3
	bucket    string
	keyPrefix string
}

var _ TestStatsReporter = (*S3Reporter)(nil)

// NewS3Reporter creates a new S3Reporter that puts test results in the given bucket using the keyPrefix.
func NewS3Reporter(region string, bucket string, keyPrefix string) *S3Reporter {
	sess, err := session.NewSession(&aws.Config{
		Region: aws.String(region),
	})
	if err != nil {
		fmt.Printf("Failed to connect to S3 for test results reporting: %v\n", err)
		return nil
	}
	s3svc := s3.New(sess)
	return &S3Reporter{
		s3svc:     s3svc,
		bucket:    bucket,
		keyPrefix: keyPrefix,
	}

}

// ReportCommand uploads the results of running a command to S3
func (r *S3Reporter) ReportCommand(stats TestCommandStats) {
	byts, err := json.Marshal(stats)
	if err != nil {
		fmt.Printf("Failed to serialize report for upload to S3: %v: %v\n", stats, err)
		return
	}
	name := resource.NewUniqueHex(fmt.Sprintf("%v-", time.Now().UnixNano()), -1, 10)
	_, err = r.s3svc.PutObject(&s3.PutObjectInput{
		Bucket: aws.String(r.bucket),
		Key:    aws.String(path.Join(r.keyPrefix, name)),
		Body:   bytes.NewReader(byts),
		ACL:    aws.String(s3.ObjectCannedACLBucketOwnerFullControl),
	})
	if err != nil {
		fmt.Printf("Failed to upload test command report to S3: %v\n", err)
		return
	}
}
